/*
ESD-Guard mit ESP8266
[c]2020 DCR Dirk Conrad
*/

#include <Ticker.h>
#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <EEPROM.h>
#include <SPI.h>
#include <ESP8266httpUpdate.h>

Ticker tTimer;
MDNSResponder mDNS;
ESP8266WebServer oWebServer(80);

// ####################################################################################################
// ### DEFINITIONEN                                                                                 ###
// ####################################################################################################

#define USART_BUFFER_SIZE       70
#define NTP_PACKET_SIZE         48

// ####################################################################################################
// ### DEFINITIONEN EEPROM                                                                          ###
// ####################################################################################################

// EPROM Größen
#define EE_SIZE               4095
#define EE_DEVICE_NAME_LEN      32
#define EE_NTP_NAME_LEN         32 // Adressraum für 3 Namen
#define EE_PIN_LEN               7
#define EE_SSID_SIZE            33
#define EE_PASSWORD_SIZE        64
#define EE_FIXED_IPS_LEN        17
#define EE_EVENT_LOG_SIZE      240 // Anzahl der zu loggenden Ereignispakete

// EPROM Adressen
#define EE_DEVICE_NAME_ADR       0 // 32 Bytes für den Gerätenamen inkl. ChkSum
#define EE_NTP_NAME_ADR         32 // 96 Bytes für 3 NTP Namen inkl. ChkSum
#define EE_TIME_ZONE_ADR       128 // 2 Bytes für Zeitzone inkl. ChkSum
#define EE_DST_ADR             130 // 2 Bytes für Sommerzeit inkl. ChkSum
#define EE_PIN_ADR             132 // 7 Bytes für PIN inkl. ChkSum
#define EE_SSID_ADR            139 // 33 Bytes für SSID inkl. ChkSum
#define EE_PASSWORD_ADR        172 // 64 Bytes für PASSWORD inkl. ChkSum
#define EE_FIXED_IPS_ADR       236 // 17 Bytes für 4 IP-Adressen inkl. ChkSum
#define EE_EVENT_PNT_ADR       253 // 2 Bytes für Pointer inkl. ChkSum
#define EE_EVENT_LOG_ADR       255 // 16 Bytes je Ereignespaket inkl. ChkSum

// ####################################################################################################
// ### GLOBALE VARIABLEN                                                                            ###
// ####################################################################################################

char sVersion[] = { "1.20.02.07" };
String sImgPath = "http://www.contronik.net/download/esd-guard/";
unsigned char ucPinAM2302 = 2;
unsigned char ucPinLEDgn = 15;
unsigned char ucPinLEDye = 16;
unsigned char ucPinLEDbl = 5;
unsigned char ucPinLEDrd = 4;
int iLEDrdCnt = 0;
int iLEDblCnt = 0;
int iLEDyeCnt = 0;
int iLEDrdSum = 0;
int iLEDblSum = 0;
int iLEDyeSum = 0;
unsigned char ucLEDrdLock = 1;
unsigned char ucLEDblLock = 1;
unsigned char ucLEDyeLock = 1;
int iLEDCnt = 0;

unsigned char ucPinSCLK = 14;
unsigned char ucPinMISO = 12;
unsigned char ucPinMOSI = 13;
unsigned char ucPinCS = 0;
int iRH = 0;
int iTM = 0;
unsigned char ucAM2302Busy = 0;
char cBuffer[USART_BUFFER_SIZE];
unsigned char ucBufferPointer = 0;
unsigned char ucOffsetPointer = 0;
int iTimerDiv = 0;
unsigned char ucReadAm2302Lock = 0;
unsigned char ucResetLEDLock = 0;
unsigned char ucNTPTimeLock = 0;
unsigned char ucWriteEventLock = 0;
unsigned char ucData1;
unsigned char ucData2;
int iADCValue;
int iADCDiv = 0;
int iEventADC = 255;
int iMaxValue = 0;
int iUSBVoltage = 0;

// Netzwerk
char sSSID[EE_SSID_SIZE];
char sPassword[EE_PASSWORD_SIZE];
char sHostname[64];
char sDeviceName[EE_DEVICE_NAME_LEN];
char sLogString[42];
unsigned char ucWebServerStarted = 0;
String sHTML = "";
String sPIN = "";
IPAddress ipAdr (127, 0, 0, 1);
IPAddress ipGateway (127, 0, 0, 2);
IPAddress ipSubnetMask (255, 255, 255, 0);
IPAddress ipDNS (8, 8, 8, 8);

// Variablen für Uhrzeit und Datum
struct RealTimeClock {
  unsigned char Second  = 0;            // Sekunde
  unsigned char Minute  = 0;            // Minute
  unsigned char Hour    = 0;            // Stunde
  unsigned char Day     = 1;            // Tag
  unsigned char Month   = 0;            // Monat
  unsigned int  Year    = 2000;         // Jahr
  unsigned char WeekDay = 6;            // Wochentag
  unsigned char DST =     0;            // Sommerzeit
};
RealTimeClock RTC;
char sWeekName[][7] = { "Su", "Mo", "Tu", "We", "Th", "Fr", "Sa" };
unsigned char ucDaysInMonth[] = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
unsigned char ucIsTimeOK = 0;

// ####################################################################################################
// ### INITIALISIERUNG                                                                              ###
// ####################################################################################################

void setup () {
int iCnt;
String sMacAdr;
unsigned char ucData;
unsigned char ucSum;

  // PWM
  analogWriteRange (255);
  analogWriteFreq (1000);

  // Pin-Mode
  pinMode (ucPinAM2302, INPUT);
  pinMode (ucPinLEDgn, OUTPUT);
  pinMode (ucPinLEDye, OUTPUT);
  pinMode (ucPinLEDbl, OUTPUT);
  pinMode (ucPinLEDrd, OUTPUT);
  analogWrite (ucPinLEDgn, 0);
  digitalWrite (ucPinLEDye, 0);
  analogWrite (ucPinLEDbl, 0);
  analogWrite (ucPinLEDrd, 0);
  pinMode (ucPinSCLK, OUTPUT);
  pinMode (ucPinMISO, INPUT);
  pinMode (ucPinMOSI, OUTPUT);
  pinMode (ucPinCS, OUTPUT);
  digitalWrite (ucPinCS, 1);
  digitalWrite (ucPinSCLK, 1);

  // RS232
  Serial.begin (115200, SERIAL_8N1);
  Serial.flush ();

  // SPI
  SPI.begin ();
  SPI.setDataMode (SPI_MODE0);
  SPI.setFrequency (8000000);

  // WIFI
  sprintf (sSSID, "");
  sprintf (sPassword, "PASSWORD");
  WiFi.mode (WIFI_STA);
  sMacAdr = WiFi.macAddress();
  sprintf (sHostname, "ESD-Guard-%s", sMacAdr.c_str());
  for (iCnt = 0; iCnt < strlen(sHostname); iCnt++) {
    if (sHostname[iCnt] == ':') {
      sHostname[iCnt] = '-';
    }
  }
  WiFi.hostname (sHostname);

  // EEPROM
  EEPROM.begin (EE_SIZE);

  // Timer
  tTimer.attach_ms (1, Timer);

  // Feste IP-Adressen aus EEPROM lesen
  ucSum = 0xAA;
  for (iCnt = 0; iCnt < 4; iCnt++) {
    ipAdr[iCnt] = EEPROM.read (EE_FIXED_IPS_ADR + iCnt + 0);
    ucSum += ipAdr[iCnt];
  }
  for (iCnt = 0; iCnt < 4; iCnt++) {
    ipGateway[iCnt] = EEPROM.read (EE_FIXED_IPS_ADR + iCnt + 4);
    ucSum += ipGateway[iCnt];
  }
  for (iCnt = 0; iCnt < 4; iCnt++) {
    ipSubnetMask[iCnt] = EEPROM.read (EE_FIXED_IPS_ADR + iCnt + 8);
    ucSum += ipSubnetMask[iCnt];
  }
  for (iCnt = 0; iCnt < 4; iCnt++) {
    ipDNS[iCnt] = EEPROM.read (EE_FIXED_IPS_ADR + iCnt + 12);
    ucSum += ipDNS[iCnt];
  }
  ucData = EEPROM.read (EE_FIXED_IPS_ADR + iCnt + 16);
  if (ucSum == ucData && ipAdr[0] != 0 && ipAdr[1] != 0 && ipAdr[2] != 0 && ipAdr[3] != 0) {
    WiFi.config (ipAdr, ipDNS, ipGateway, ipSubnetMask);
    // Neu verbinden
    ucSum = 0xAA;
    for (iCnt = 0; iCnt < EE_SSID_SIZE -1; iCnt++) {
      ucData = EEPROM.read (EE_SSID_ADR + iCnt);
      sSSID [iCnt] = ucData;
      ucSum += ucData;
    }
    sSSID[iCnt] = 0;
    ucData = EEPROM.read (EE_SSID_ADR + iCnt);
    if (ucSum == ucData) {
      ucSum = 0xAA;
      for (iCnt = 0; iCnt < EE_PASSWORD_SIZE -1; iCnt++) {
        ucData = EEPROM.read (EE_PASSWORD_ADR + iCnt);
        sPassword [iCnt] = ucData;
        ucSum += ucData;
      }
      sPassword[iCnt] = 0;
      ucData = EEPROM.read (EE_PASSWORD_ADR + iCnt);
      if (ucSum == ucData) {
        WiFi.begin (sSSID, sPassword);
      }
    }      
  }
}

// ####################################################################################################
// ### HAUPTPROGRAMM                                                                                ###
// ####################################################################################################

void loop () {
  // Kommunikation über USB
  Serial_Commands ();

 // Zähler Locks
  if (!ucResetLEDLock && iTimerDiv == 0) {
    ucLEDrdLock = 0;
    ucLEDblLock = 0;
    ucLEDyeLock = 0;
    ucResetLEDLock = 1;
  }

  // In größeren Abständen Temperatur und Feuchte holen
  if (!ucReadAm2302Lock && iTimerDiv == 250 && (RTC.Second == 1 || RTC.Second == 31)) {
    Read_AM2302 ();
    ucReadAm2302Lock = 1;
  }

  // Zeit im Internet synchronisieren
  if (!ucNTPTimeLock && iTimerDiv == 500) {
    if (RTC.Second == 5 && ucIsTimeOK == 0) {
      Get_Time_From_NTP (0);
    }
    else if (RTC.Second == 0 && RTC.Minute == 30) {
      Get_Time_From_NTP (0);
    }
    ucNTPTimeLock = 1;
  }

  // Event-Log
  if (!ucWriteEventLock && iTimerDiv == 750) {
    if (RTC.Second == 0) {
      Write_Event_Log ();
    }
    ucWriteEventLock = 1;
  }

   if (iADCDiv == 0) {
    // ADC N-CH
    Read_ADC_0 ();
    iADCValue = Read_ADC_0 ();
    iADCValue >>= 3;
    iADCValue -= 56;
    if (iADCValue < 0) { iADCValue = 0; }
    if (iADCValue > 400) { iADCValue = 400; }
    if (iLEDrdCnt <= iADCValue) {
      iLEDrdCnt += iADCValue;
      iLEDrdCnt >>= 1;
      if (!ucLEDrdLock) {
        if (iLEDrdCnt > 10)  { ucLEDrdLock = 1; iLEDrdSum++; }
        if (iLEDrdCnt > 50)  { ucLEDrdLock = 1; iLEDrdSum++; }
        if (iLEDrdCnt > 100) { ucLEDrdLock = 1; iLEDrdSum++; }
        if (iLEDrdCnt > 200) { ucLEDrdLock = 1; iLEDrdSum++; }
        if (iLEDrdCnt > 300) { ucLEDrdLock = 1; iLEDrdSum++; }
        if (iLEDrdSum > 255) { iLEDrdSum = 255; }
      }
    }
    // ADC P-CH
    Read_ADC_1 ();
    iADCValue = 4095 - Read_ADC_1 ();
    iADCValue >>= 3;
    iADCValue -= 56;
    if (iADCValue < 0) { iADCValue = 0; }
    if (iADCValue > 400) { iADCValue = 400; }
    if (iLEDblCnt <= iADCValue) {
      iLEDblCnt += iADCValue;
      iLEDblCnt >>= 1;
      if (!ucLEDblLock) {
        if (iLEDblCnt > 10)  { ucLEDblLock = 1; iLEDblSum++; }
        if (iLEDblCnt > 50)  { ucLEDblLock = 1; iLEDblSum++; }
        if (iLEDblCnt > 100) { ucLEDblLock = 1; iLEDblSum++; }
        if (iLEDblCnt > 200) { ucLEDblLock = 1; iLEDblSum++; }
        if (iLEDblCnt > 300) { ucLEDblLock = 1; iLEDblSum++; }
        if (iLEDblSum > 255) { iLEDblSum = 255; }
      }
    }
  }

  if (iADCDiv == 1) {
    // USB Spannung
    Read_ADC_3 ();
    iADCValue = Read_ADC_3 ();
    iUSBVoltage = ((double)iADCValue / 6.2);
    // Mux wechseln
    Read_ADC_2 ();
  }

  if (iADCDiv > 1) {
    // ADC Peak
    iADCValue = Read_ADC_2 ();
    iADCValue *= 0.08;
    if (iADCValue < 10) { iADCValue = 0; }
    iEventADC += iADCValue;
    iEventADC >>= 1;
    if (iMaxValue < iEventADC) {
      iMaxValue = iEventADC;
      if (iLEDyeCnt < iMaxValue) {
        iLEDyeCnt = iMaxValue;
        if (!ucLEDyeLock) {
          if (iLEDrdCnt > 10 || iLEDblCnt > 10) {
            if (iLEDyeCnt > 20)  { ucLEDyeLock = 1; iLEDyeSum++; }
            if (iLEDyeCnt > 40)  { ucLEDyeLock = 1; iLEDyeSum++; }
            if (iLEDyeCnt > 60)  { ucLEDyeLock = 1; iLEDyeSum++; }
            if (iLEDyeCnt > 80)  { ucLEDyeLock = 1; iLEDyeSum++; }
            if (iLEDyeCnt > 100) { ucLEDyeLock = 1; iLEDyeSum++; }
            if (iLEDyeSum > 255) { iLEDyeSum = 255; }
          }
        }
      }
    }
    else {
      if (iMaxValue > 0) {
        iMaxValue--;
      }
    }
  }

  iADCDiv++;
  if (iADCDiv > 100) {
    iADCDiv = 0;
  }

  // Web-Server
  if (ucWebServerStarted == 0) {
    if (WiFi.status () == WL_CONNECTED) {
      if (mDNS.begin ("esd-guard", WiFi.localIP())) {
        ucWebServerStarted = 1;
        oWebServer.on ("/", Build_Website);
        oWebServer.begin ();
      }
    }
  }
  else {
    oWebServer.handleClient ();
  }
}

// ########## Webseite zusammenbauen ##########

void Build_Website () {
int iCnt;
int iRefCnt;
int iStartPnt;
int iCurCnt;
int iTimeHr;
int iTimeDy;
char sTmp[32];
unsigned char ucPINOK = 0;
String sAct;
String sImg;
String sUpd;
String sTitle;

ipAdr = WiFi.localIP ();
ipGateway = WiFi.gatewayIP ();
ipSubnetMask = WiFi.subnetMask ();

  oWebServer.sendContent ("HTTP/1.1 200 OK\n");

  sPIN = oWebServer.arg ("p");
  Read_Network_PIN (sTmp, 0);
  if (sPIN == String(sTmp)) {
    ucPINOK = 1;
  }

  sAct = oWebServer.arg ("a");
  if (sAct == "reset") {
    Reset_Event_Log ();
  }

  sHTML = "";
  sHTML += "Content-Type: text/html; charset=ISO-8859-1\n\n";
  sHTML += "<HTML>";
  sHTML += "<HEAD>";
  sHTML += "<TITLE>ESD-Guard Webinterface</TITLE>";
  sHTML += "<META NAME=\"viewport\" content=\"width=500\">";
  sHTML += "<STYLE type=\"text/css\">";
  sHTML += "<!--";
  sHTML += "body, td, font, input, textarea, select {font-family: \"Lucida Sans Typewriter\", \"Lucida Console\", monaco, \"Bitstream Vera Sans Mono\", monospace;}";
  sHTML += "font, th, td, p {font-family: \"Lucida Sans Typewriter\", \"Lucida Console\", monaco, \"Bitstream Vera Sans Mono\", monospace;} a:link {text-decoration: none;}";
  sHTML += "a:visited {text-decoration: none;} a:hover {text-decoration: underline;}";
  sHTML += "hr {border-top: 1px solid #808080; border-bottom: 1px solid #808080;}";
  sHTML += "th {color: #000000; font-size: 14px; font-weight: bold;}";
  sHTML += "td {color: #000000; font-size: 14px;}";
  sHTML += "-->";
  sHTML += "</STYLE>";
  sHTML += "</HEAD>";
  oWebServer.sendContent (sHTML);

  sUpd = oWebServer.arg ("u");
  if (sUpd != "") {
    OTA_Update (sUpd);
    return;
  }
  
  sHTML = "";
  sHTML += "<BODY BGCOLOR=#F0F0F0 TEXT=#000000 LINK=#000040 VLINK=#000040 ALINK=#0000A0>";
  sHTML += "<TABLE BGCOLOR=#E8E8E8 CELLPADDING=0 CELLSPACING=10 ALIGN=center BORDER=0>";
  sHTML += "<TR>";
  sHTML += "<TD ALIGN=left>";
  sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 ALIGN=center WIDTH=480 BORDER=0>";
  sHTML += "<TR>";
  sHTML += "<TD ALIGN=center VALIGN=center>";
  sHTML += "<DIV><IMG SRC=" + sImgPath + "IMG_ESD-Guard.gif></DIV>";
  sHTML += "<DIV>";
  sHTML += "<B>Version " + String(sVersion) + "</B>";
  if (ucPINOK) {
    sHTML += " <FORM ACTION=\"\" METHOD=get><INPUT TYPE=hidden NAME=p VALUE=\"" + String(sPIN) + "\"><INPUT TYPE=hidden NAME=u VALUE=1><INPUT TYPE=submit VALUE=\"FIRMWARE UPDATE\"></FORM>";
  }
  sHTML += "</DIV>"; 
  sHTML += "<HR>";
  sHTML += "<DIV style=\"font-size:12px; color:#000000;\">";
  sHTML += "<B>" + String(sHostname) + "</B> @ " + String(ipAdr[0], DEC) + "." + String(ipAdr[1], DEC) + "." + String(ipAdr[2], DEC) + "." + String(ipAdr[3], DEC);
  sHTML += "</DIV>";
  sHTML += "</TD>";
  sHTML += "</TR>";
  sHTML += "</TABLE>";
  sHTML += "</TD>";
  sHTML += "</TR>";
  sHTML += "<TR>";
  sHTML += "<TD ALIGN=left>";
  sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 ALIGN=center WIDTH=480 BORDER=0>";
  sHTML += "<TR>";
  sHTML += "<TD ALIGN=left COLSPAN=4><B>Device Name</B>";
  sHTML += "</TD>";
  sHTML += "</TR>";
  Read_Device_Name (0);
  sHTML += "<TR><TD ALIGN=left COLSPAN=4 BGCOLOR=#FFFFFF><B>" + String(sDeviceName) + "</B></TD></TR>";
  sprintf (sTmp, "%s %0*d-%0*d-%0*d %0*d:%0*d:%0*d\r\n", sWeekName[RTC.WeekDay], 4, RTC.Year, 2, RTC.Month + 1, 2, RTC.Day, 2, RTC.Hour, 2, RTC.Minute, 2, RTC.Second);
  sHTML += "<TR><TD ALIGN=left VALIGN=center><IMG SRC=" + sImgPath + "IMG_Date.gif></TD><TD ALIGN=left><IMG SRC=" + sImgPath + "IMG_Temp.gif></TD><TD ALIGN=left><IMG SRC=" + sImgPath + "IMG_RH.gif></TD><TD ALIGN=left><B>USB</B></TD></TR>";
  sHTML += "<TR><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String(sTmp) + "</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String((double)iTM / 10.0, 1) + "&deg;C</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String((double)iRH / 10.0, 1) + "%</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String((double)iUSBVoltage / 100.0, 2) + " V</B><TD></TR>";
  sHTML += "</TABLE>";
  sHTML += "</TD>";
  sHTML += "</TR>";
  sHTML += "<TR>";
  sHTML += "<TD ALIGN=left>";
  oWebServer.sendContent (sHTML);

  if (ucPINOK) {
    sHTML = "";
    sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 ALIGN=center WIDTH=480 BORDER=0>";
    sHTML += "<TR>";
    sHTML += "<FORM ACTION=\"\" METHOD=get><TD ALIGN=center><INPUT TYPE=submit VALUE=\"UPDATE\"><INPUT TYPE=hidden NAME=a VALUE=reload><INPUT TYPE=hidden NAME=p VALUE=\"" + String(sPIN) + "\"></TD></FORM>";
    sHTML += "<FORM ACTION=\"\" METHOD=get><TD ALIGN=center><INPUT TYPE=submit VALUE=\"RESET\"><INPUT TYPE=hidden NAME=a VALUE=reset><INPUT TYPE=hidden NAME=p VALUE=\"" + String(sPIN) + "\"></TD></FORM>";
    sHTML += "</TR>";
    sHTML += "</TABLE>";
    sHTML += "</TD>";
    sHTML += "</TR>";
    sHTML += "<TR>";
    sHTML += "<TD ALIGN=left>";
    sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 WIDTH=480 ALIGN=center STYLE=\"border: 2px solid #808080;\">";
    sHTML += "<TR>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><B>#</B></TD>";
    sHTML += "<TD ALIGN=left BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_Date.gif></TD>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_N.gif></TD>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_P.gif></TD>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_E.gif></TD>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_Temp.gif></TD>";
    sHTML += "<TD ALIGN=center BGCOLOR=#D8D8D8><IMG SRC=" + sImgPath + "IMG_RH.gif></TD>";
    sHTML += "</TR>";
    oWebServer.sendContent (sHTML);

    iCurCnt = Read_Event_Log_Pointer (1);
    iStartPnt = iCurCnt;
    iCnt = 1;
    iRefCnt = 0;
    sHTML = "";
    do {
      iCurCnt--;
      if (iCurCnt < 0) {
        iCurCnt = (EE_EVENT_LOG_SIZE - 1);
      }
      Read_Event_Log (0, iCurCnt);
      sLogString[19] = 0;
      sLogString[23] = 0;
      sLogString[27] = 0;
      sLogString[31] = 0;
      sLogString[36] = 0;
      if (sLogString[0] == '#') {
        break;
      }
      sHTML += "<TR><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String(iCnt) + "</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String(sLogString);
      if (sLogString[20] == '0' && sLogString[21] == '0' && sLogString[22] == '0') {
        sHTML += "</B></TD><TD ALIGN=center><B>";
      }
      else {
        sHTML += "</B></TD><TD ALIGN=center BGCOLOR=#FFD0D0><B>";
      }
      sHTML += String(sLogString + 20);
      if (sLogString[24] == '0' && sLogString[25] == '0' && sLogString[26] == '0') {
        sHTML += "</B></TD><TD ALIGN=center><B>";
      }
      else {
        sHTML += "</B></TD><TD ALIGN=center BGCOLOR=#D0D0FF><B>";
      }
      sHTML += String(sLogString + 24);
      if (sLogString[28] == '0' && sLogString[29] == '0' && sLogString[30] == '0') {
        sHTML += "</B></TD><TD ALIGN=center><B>";
      }
      else {
        sHTML += "</B></TD><TD ALIGN=center BGCOLOR=#FFFFD0><B>";
      }
      sHTML += String(sLogString + 28);
      sHTML += "</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String(sLogString + 32) + "&deg;C</B></TD><TD ALIGN=center BGCOLOR=#FFFFFF><B>" + String(sLogString + 37) + "%</B></TD></TR>";
      iRefCnt++;
      if (iRefCnt == 10) {
        iRefCnt = 0;
        oWebServer.sendContent (sHTML);
        sHTML = "";
      }
      iCnt++;
    } while (iCurCnt != iStartPnt);

    sHTML += "</TABLE>";
    oWebServer.sendContent (sHTML);

  }
  else {
    // PIN-Eingabe
    sHTML = "";
    sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 ALIGN=center WIDTH=480 BORDER=0>";
    sHTML += "<FORM ACTION=\"\" METHOD=get>";
    sHTML += "<TR>";
    sHTML += "<TD ALIGN=center>";
    if (sPIN != "") {
      sHTML += "<FONT COLOR=FF0000><B>Last PIN was wrong!</B></FONT><BR>";
    }
    sHTML += "Enter PIN";
    sHTML += "</TD>";
    sHTML += "</TR>";
    sHTML += "<TR><TD ALIGN=center><INPUT TYPE=text SIZE=15 NAME=p><INPUT TYPE=hidden NAME=a VALUE=pin></TD></TR>";
    sHTML += "<TR><TD ALIGN=center><INPUT TYPE=submit VALUE=\"LOGIN\"></TD></TR>";
    sHTML += "</FORM>";
    sHTML += "</TABLE>";
    oWebServer.sendContent (sHTML);
  }

  sHTML = "";
  sHTML += "</TD>";
  sHTML += "</TR>";
  sHTML += "</TABLE>";
  sHTML += "</BODY>";
  sHTML += "</HTML>";
  oWebServer.sendContent (sHTML);
}

// ########## Firmware Update ##########

void OTA_Update (String sUpd) {
const char* sHost = "www.contronik.net";
const char* sURL = "/download/esd-guard/ESD-Guard_Firmware.ino.nodemcu.bin";
String sLine = "-";
unsigned long ulTimeOut;
unsigned char isClientOK = 1;
int iTry;
char sDate[10];

  sHTML = "";
  sHTML += "<BODY BGCOLOR=#F0F0F0 TEXT=#000000 LINK=#000040 VLINK=#000040 ALINK=#0000A0>";
  sHTML += "<TABLE BGCOLOR=#E8E8E8 CELLPADDING=0 CELLSPACING=10 ALIGN=center BORDER=0><TR><TD ALIGN=left>";
  sHTML += "<B><U>Firmware Update</U></B><BR><BR>";

  if (sUpd == "1") {
    sHTML += "Flash size (bytes): " + String (ESP.getFreeSketchSpace ()) + "<BR><BR>";
    sprintf (sDate, "%c%c.%c%c.\'%c%c", sVersion[8], sVersion[9], sVersion[5], sVersion[6], sVersion[2], sVersion[3]);
    sHTML += "Current version: " + String (sVersion) + " (" + String (sDate) + ")<BR><BR>";

    WiFiClient client;
    if (!client.connect (sHost, 80)) {
      isClientOK = 0;
    }

    if (isClientOK) {
      client.print (String ("GET ") + sURL + " HTTP/1.1\r\n" + "Host: " + sHost + "\r\n" + "Connection: close\r\n\r\n");
  
      ulTimeOut = millis();
      while (client.available () == 0) {
        if (millis() - ulTimeOut > 3000) {
          sHTML += "No version information available.<BR>";
          client.stop ();
          isClientOK = 0;
          break;
        }
      }
    }

    sHTML += "Version on server:<BR>";
    
    if (isClientOK) {
      iTry = 20;
      while (sLine != "\n") {
        iTry--;
        if (iTry == 0) {
          break;
        }
        sLine = client.readStringUntil ('\r');
        sHTML += sLine + "<BR>";
      }
    }
  }
 
  if (sUpd == "2") {
    sHTML += "Updating ... please wait!<BR><BR>";
  }
  sHTML += "<HR>";
  sHTML += "<TABLE CELLPADDING=1 CELLSPACING=3 ALIGN=center WIDTH=480 BORDER=0>";
   sHTML += "<TR>";
  if (sUpd != "2") {
    sHTML += "<TD ALIGN=center><FORM ACTION=\"\" METHOD=get><INPUT TYPE=hidden NAME=p VALUE=\"" + String(sPIN) + "\"><INPUT TYPE=hidden NAME=u VALUE=2><INPUT TYPE=submit VALUE=\"FLASH\"></FORM></TD>";
  }
  sHTML += "<TD ALIGN=center><FORM ACTION=\"\" METHOD=get><INPUT TYPE=hidden NAME=p VALUE=\"" + String(sPIN) + "\"><INPUT TYPE=submit VALUE=\"HOME\"></FORM></TD>";
  sHTML += "</TR>";
  sHTML += "</TABLE>";
  oWebServer.sendContent (sHTML);

  if (sUpd == "2") {
    t_httpUpdate_return ret = ESPhttpUpdate.update ("www.contronik.net", 80, "/download/esd-guard/ESD-Guard_Firmware.ino.nodemcu.bin");
    if (ret != HTTP_UPDATE_OK) {
        oWebServer.sendContent ("<BR>Update failed.");
    }
  }

  sHTML = ""; 
  sHTML += "</TD></TR></TABLE>";
  sHTML += "</BODY>";
  sHTML += "</HTML>";
  oWebServer.sendContent (sHTML);
}

// ########## Timer ##########

void Timer () {
  // Auslesen des Temp./RH Sensors darf nicht gestört werden
  if (ucAM2302Busy) { return; }

  // TimerDiv zählt ms
  iTimerDiv++;
  // Teiler Uhrzeit
  if (iTimerDiv >= 1000) {
    iTimerDiv = 0;
    ucReadAm2302Lock = 0;
    ucResetLEDLock = 0;
    ucNTPTimeLock = 0;
    ucWriteEventLock = 0;
    RTC.Second++;
    if (RTC.Second >= 60) {
      RTC.Second = 0;
      RTC.Minute++;
      if (RTC.Minute >= 60) {
        RTC.Minute = 0;
        RTC.Hour++;
        if (RTC.Hour >= 24) {
          RTC.Hour = 0;
          RTC.WeekDay++;
          if (RTC.WeekDay > 6) {
            RTC.WeekDay = 0;
          }
          RTC.Day++;
          if (RTC.Day > ucDaysInMonth[RTC.Month]) {
            RTC.Day = 1;
            RTC.Month++;
            if (RTC.Month > 11) {
              RTC.Month = 0;
              RTC.Year++;
            }
          }
        }
      }
    }
  }

  // PWM für rote und blaue LEDs
  analogWrite (ucPinLEDrd, iLEDrdCnt);
  analogWrite (ucPinLEDbl, iLEDblCnt);

  // Anzeige gelbe LED
  if (iLEDyeCnt) {
    digitalWrite (ucPinLEDye, 1);
  }
  else {
    digitalWrite (ucPinLEDye, 0);
  }

  // Verzögertes Abklingen der LEDs
  iLEDCnt++;
  if (iLEDCnt == 10) {
    iLEDCnt = 0;
    if (iLEDyeCnt > 0) {
      iLEDyeCnt--;
    }
    if (iLEDrdCnt >= 2) {
      iLEDrdCnt -=2;
    }
    if (iLEDrdCnt < 2) {
      iLEDrdCnt = 0;
    }
    if (iLEDblCnt >= 2) {
      iLEDblCnt -=2;
    }
    if (iLEDblCnt < 2) {
      iLEDblCnt = 0;
    }
  }

  // Grüne LED zeigt den WEB-Server Status
  if (iTimerDiv == 0) {
    if (ucWebServerStarted) {
      analogWrite (ucPinLEDgn, 128);
    }
    else {
      analogWrite (ucPinLEDgn, 255);
    }
  }
  if (iTimerDiv == 20 && !ucWebServerStarted) {
    analogWrite (ucPinLEDgn, 0);
  }
}

// ########## Kommandos von der seriellen Schnittstelle ##########

void Serial_Commands () {
char sData;
char sTmp[32];

  if (Serial.available() > 0) {
    sData = Serial.read ();

    // Return empfangen?
    if (sData == '\r') {
      cBuffer[ucBufferPointer] = 0;
      ucBufferPointer = 0;

      // Hilfe
      if (Compare ("HELP")) {
        Serial.printf ("ID .......... Device identification\r\n");
        Serial.printf ("SCAN ........ Network scan\r\n");
        Serial.printf ("NS .......... Read network status\r\n");
        Serial.printf ("HST ......... Read host name\r\n");
        Serial.printf ("LIP ......... Read local IP\r\n");
        Serial.printf ("GIP ......... Read gateway IP\r\n");
        Serial.printf ("SUB ......... Read subnet mask\r\n");
        Serial.printf ("DNS ......... Read DNS\r\n");
        Serial.printf ("MAC ......... Read MAC\r\n");
        Serial.printf ("WEB ......... Get web server status\r\n");
        Serial.printf ("RID.......... Read SSID\r\n");
        Serial.printf ("WID [name] .. Write SSID\r\n");
        Serial.printf ("PW [pw] ..... Set password\r\n");
        Serial.printf ("RPIN ........ Read webinterface pin\r\n");
        Serial.printf ("WPIN [pw] ... Write webinterface pin\r\n");
        Serial.printf ("CON ......... Connect to station\r\n");
        Serial.printf ("FIX [ip,..].. Set local IP, gateway IP, subnet mask and DNS\r\n");
        Serial.printf ("USB.......... Read USB voltage\r\n");
        Serial.printf ("TM........... Read temperature\r\n");
        Serial.printf ("RH........... Read relative humidity\r\n");
        Serial.printf ("RV .......... Read negative static raw value\r\n");
        Serial.printf ("RS .......... Read negative static sum counter\r\n");
        Serial.printf ("BV .......... Read positive static raw value\r\n");
        Serial.printf ("BS .......... Read positive static sum counter\r\n");
        Serial.printf ("YV .......... Read ESD event raw value\r\n");
        Serial.printf ("YS .......... Read ESD event sum counter\r\n");
        Serial.printf ("TIME ........ Read time\r\n");
        Serial.printf ("TSW [wday]....Set week day\r\n");
        Serial.printf ("TSY [year]....Set year\r\n");
        Serial.printf ("TSO [month]...Set month\r\n");
        Serial.printf ("TSD [day].....Set day\r\n");
        Serial.printf ("TSH [hour]....Set hour\r\n");
        Serial.printf ("TSM [min].....Set minute\r\n");
        Serial.printf ("TSS [sec].....Set second\r\n");
        Serial.printf ("NTP ......... Get time from NTP-Server\r\n");
        Serial.printf ("RNTPn ....... Read NTP name 1-3\r\n");
        Serial.printf ("WNTPn [name]. Write NTP name 1-3\r\n");
        Serial.printf ("RTZ ......... Read time zone\r\n");
        Serial.printf ("WTZ [hour]... Write time zone\r\n");
        Serial.printf ("RDST ........ Read daylight saving time\r\n");
        Serial.printf ("WDST [0/1]... Write daylight saving time\r\n");
        Serial.printf ("RDN ......... Read device name\r\n");
        Serial.printf ("WDN [name]... Write device name\r\n");
        Serial.printf ("REL [n]...... Read event log\r\n");
        Serial.printf ("PNT ......... Read event log pointer value\r\n");
        Serial.printf ("RST ......... Reset event log\r\n");
      }
      // Hardware identifizieren
      else if (Compare ("ID")) {
        Serial.printf ("ESD-Guard Version %s\r\n", sVersion);
      }
      // Netzwerk Scan
      else if (Compare ("SCAN")) {
        Network_Scan ();
      }
      // Netzwerk Status
      else if (Compare ("NS")) {
        switch (WiFi.status ()) {
          case  0: Serial.printf ("IDLE_STATUS\r\n"); break;
          case  1: Serial.printf ("NO_SSID_AVAILABLE\r\n"); break;
          case  3: Serial.printf ("CONNECTED\r\n"); break;
          case  4: Serial.printf ("CONNECT_FAILED\r\n"); break;
          case  6: Serial.printf ("DISCONNECTED\r\n"); break;
          default: Serial.printf ("UNDEFINED\r\n"); break;
        }
      }
      // Hostname
      else if (Compare ("HST")) {
        Serial.println (WiFi.hostname ());
      }
      // Local IP
      else if (Compare ("LIP")) {
        Serial.println (WiFi.localIP ());
      }
      // Gateway IP
      else if (Compare ("GIP")) {
        Serial.println (WiFi.gatewayIP ());
      }
      // Subnet Mask
      else if (Compare ("SUB")) {
        Serial.println (WiFi.subnetMask ());
      }
      // DNS
      else if (Compare ("DNS")) {
        Serial.println (WiFi.dnsIP ());
      }
      // MAC
      else if (Compare ("MAC")) {
        Serial.println (WiFi.macAddress ());
      }
      // Web-Server
      else if (Compare ("WEB")) {
        Serial.printf ("%d\r\n", ucWebServerStarted);
      }
      // Read SSID
      else if (Compare ("RID")) {
        Serial.println (WiFi.SSID ());
      }
      // SSID setzen
      else if (Compare ("WID")) {
        Network_SSID ((cBuffer + ucOffsetPointer));
      }
      // Passwort setzen
      else if (Compare ("PW")) {
        Network_Password ((cBuffer + ucOffsetPointer));
      }
      // PIN lesen
      else if (Compare ("RPIN")) {
        Read_Network_PIN (sTmp, 1);
      }
      // PIN setzen
      else if (Compare ("WPIN")) {
        Write_Network_PIN ((cBuffer + ucOffsetPointer));
      }
      // Netzwerk verbinden DHCP
      else if (Compare ("CON")) {
        Network_Connect ();
      }
      // Netzwerk auf feste IPs setzen
      else if (Compare ("FIX")) {
        Fixed_IPs ();
      }
      // USB Spannung
      else if (Compare ("USB")) {
        sprintf (sTmp, "%0*d", 3, iUSBVoltage);
        Serial.printf ("%c.%c%c\r\n", sTmp[0], sTmp[1], sTmp[2]);
      }
      // Temperatur
      else if (Compare ("TM")) {
        sprintf (sTmp, "%0*d", 3, iTM);
        Serial.printf ("%c%c.%c\r\n", sTmp[0], sTmp[1], sTmp[2]);
      }
      // Relative Feuchte
      else if (Compare ("RH")) {
        sprintf (sTmp, "%0*d", 3, iRH);
        Serial.printf ("%c%c.%c\r\n", sTmp[0], sTmp[1], sTmp[2]);
      }
      // ESD Wert ROT
      else if (Compare ("RV")) {
        Serial.printf ("%0*d\r\n", 3, iLEDrdCnt);
      }
      else if (Compare ("RS")) {
        Serial.printf ("%0*d\r\n", 3, iLEDrdSum);
      }
      // ESD Wert BLAU
      else if (Compare ("BV")) {
        Serial.printf ("%0*d\r\n", 3, iLEDblCnt);
      }
      else if (Compare ("BS")) {
        Serial.printf ("%0*d\r\n", 3, iLEDblSum);
      }
      // ESD Wert GELB
      else if (Compare ("YV")) {
        Serial.printf ("%0*d\r\n", 3, iLEDyeCnt);
      }
      else if (Compare ("YS")) {
        Serial.printf ("%0*d\r\n", 3, iLEDyeSum);
      }
      // Uhrzeit lesen
      else if (Compare ("TIME")) {
        Serial.printf ("%s %0*d-%0*d-%0*d %0*d:%0*d:%0*d\r\n", sWeekName[RTC.WeekDay], 4, RTC.Year, 2, RTC.Month + 1, 2, RTC.Day, 2, RTC.Hour, 2, RTC.Minute, 2, RTC.Second);
      }
      // Uhrzeit per Cmd setzen
      else if (Compare ("TSW")) {
        RTC.WeekDay = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSY")) {
        RTC.Year = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSO")) {
        RTC.Month = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSD")) {
        RTC.Day = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSH")) {
        RTC.Hour = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSM")) {
        RTC.Minute = String_2_Value (cBuffer + ucOffsetPointer);
      }
      else if (Compare ("TSS")) {
        RTC.Second = String_2_Value (cBuffer + ucOffsetPointer);
      }
      // Gerätename lesen
      else if (Compare ("RDN")) {
        Read_Device_Name (1);
      }
      // Gerätename schreiben
      else if (Compare ("WDN")) {
        Write_Device_Name ((cBuffer + ucOffsetPointer));
      }
      // Zeit von NTP-Server holen
      else if (Compare ("NTP")) {
        Get_Time_From_NTP (1);
      }
      // NTP Server Name 1 lesen
      else if (Compare ("RNTP1")) {
        Read_NTP_Name (0, sTmp, 1);
      }
      // NTP Server Name 1 schreiben
      else if (Compare ("WNTP1")) {
        Write_NTP_Name (0, (cBuffer + ucOffsetPointer));
      }
      // NTP Server Name 2 lesen
      else if (Compare ("RNTP2")) {
        Read_NTP_Name (1, sTmp, 1);
      }
      // NTP Server Name 2 schreiben
      else if (Compare ("WNTP2")) {
        Write_NTP_Name (1, (cBuffer + ucOffsetPointer));
      }
      // NTP Server Name 3 lesen
      else if (Compare ("RNTP3")) {
        Read_NTP_Name (2, sTmp, 1);
      }
      // NTP Server Name 3 schreiben
      else if (Compare ("WNTP3")) {
        Write_NTP_Name (2, (cBuffer + ucOffsetPointer));
      }
      // Zeitzone lesen
      else if (Compare ("RTZ")) {
        Read_Time_Zone (1);
      }
      // Zeitzone schreiben
      else if (Compare ("WTZ")) {
        Write_Time_Zone ((cBuffer + ucOffsetPointer));
        Get_Time_From_NTP (0);
      }
      // DST lesen
      else if (Compare ("RDST")) {
        Read_DST (1);
      }
      // DST schreiben
      else if (Compare ("WDST")) {
        Write_DST ((cBuffer + ucOffsetPointer));
        Get_Time_From_NTP (0);
      }
      // Eine Zeile aus dem Event-Log lesen
      else if (Compare ("REL")) {
        Read_Event_Log (1, String_2_Value (cBuffer + ucOffsetPointer));
      }
      // Event-Log Pointer lesen
      else if (Compare ("PNT")) {
        Read_Event_Log_Pointer (1);
      }
      // Reset Event-Log
      else if (Compare ("RST")) {
        Reset_Event_Log ();
      }
      // Kommando nicht gefunden
      else {
        Serial.printf ("CMD not found!\r\n");
      }
    }
    else {
      if (ucBufferPointer < USART_BUFFER_SIZE-1) {
        cBuffer[ucBufferPointer] = sData;
        ucBufferPointer++;
      }
    }
  }
}

// ########## Commando vergleichen ##########

unsigned char Compare (char *sCommand) {
unsigned char ucIndex = 0;
unsigned char ucEqual = 1;
unsigned char ucUpperCase = 1;

  ucOffsetPointer = 0;
  while (sCommand[ucIndex] != '\0' && sCommand[ucIndex] != ' ') {
    if (cBuffer[ucIndex] == ' ') {
      ucUpperCase = 0;
    }
    if (cBuffer[ucIndex] > 96 && cBuffer[ucIndex] < 123 && ucUpperCase) {
      cBuffer[ucIndex] -= 32;
    }
    if (sCommand[ucIndex] != cBuffer[ucIndex]) { ucEqual = 0; }
    ucIndex++;
  }
  if ((cBuffer[ucIndex] != '\0' && cBuffer[ucIndex] != ' ')|| cBuffer[0] == '\0') {
    ucEqual = 0;
  }
  if (ucEqual) {
    if (cBuffer[ucIndex] == ' ') {
      ucEqual = 0;
      ucIndex++;
      if (cBuffer[ucIndex] != '\0') {
        ucEqual = 1;
        ucOffsetPointer = ucIndex;
      }
    }
  }
  if (ucEqual) {
    Serial.printf ("CMD:");
    Serial.printf ((char *)cBuffer);
    Serial.printf (":OK\r\n");
  }
  return ucEqual;
}

// ########## Temperatur und Feuchte lesen ##########

void Read_AM2302 () {
unsigned int uiCnt;
unsigned int uiPnt;
unsigned char ucBits;
unsigned int uiMid;
unsigned int iAM2302Resp[41];
unsigned char ucByte[5];
unsigned char ucSign;
unsigned char ucCSum;

  // Zeitkritische Operation
  ucAM2302Busy = 1;

  // AM2302 Start Signal
  pinMode (ucPinAM2302, OUTPUT);
  digitalWrite (ucPinAM2302, 0);
  delay (2);
  // AM2302 End of Start Signal
  pinMode (ucPinAM2302, INPUT);
  // Warten bis LO-Signal
  for (uiCnt=0; uiCnt < 1000; uiCnt++) {
    if (digitalRead (ucPinAM2302) == 0) {
      break;
    }
  }

  // 41 Bits lesen (1. Bit ist Sync mit 80us)
  uiPnt = 0;
  for (ucBits=0; ucBits < 41; ucBits++) {
    for (uiCnt=0; uiCnt < 1000; uiCnt++) {
      if (digitalRead (ucPinAM2302) == 1) {
        break;
      }
    }
    iAM2302Resp[uiPnt] = 0;
    for (uiCnt=0; uiCnt < 1000; uiCnt++) {
      if (digitalRead (ucPinAM2302) == 1) {
        iAM2302Resp[uiPnt]++;
      }
      else {
        break;
      }
    }
    uiPnt++;
  }

  ucAM2302Busy = 0;

  // AM2302 Auswertung
  uiMid = iAM2302Resp[0] * 0.7;
  ucByte[0] = 0;
  ucByte[1] = 0;
  ucByte[2] = 0;
  ucByte[3] = 0;
  ucByte[4] = 0;
  ucSign = 128;
  uiCnt = 0;
  ucBits = 0;
  while (uiCnt < 5) {
    ucBits++;
    if (iAM2302Resp[ucBits] > uiMid) {
      ucByte[uiCnt] |= ucSign;
    }
    ucSign >>= 1;
    if (ucSign == 0) {
      ucSign = 128;
      uiCnt++;
    }
  }
  ucCSum = ucByte[0] + ucByte[1] + ucByte[2] + ucByte[3];
  if (ucCSum == ucByte[4]) {
    iRH = (ucByte[0] << 8) + ucByte[1];
    iTM = (ucByte[2] << 8) + ucByte[3];
  }
  else {
    iRH = 0;
    iTM = 0;
  }
}

// ########## ADC AUSLESEN ##########

unsigned int Read_ADC_0 () {
  digitalWrite (ucPinCS, 0);
  ucData1 = SPI.transfer (0x00);
  ucData2 = SPI.transfer (0x00);
  digitalWrite (ucPinCS, 1);
  return (((ucData1 & 0x0F) << 8) | ucData2);
}

unsigned int Read_ADC_1 () {
  digitalWrite (ucPinCS, 0);
  ucData1 = SPI.transfer (0x08);
  ucData2 = SPI.transfer (0x00);
  digitalWrite (ucPinCS, 1);
  return (((ucData1 & 0x0F) << 8) | ucData2);
}

unsigned int Read_ADC_2 () {
  digitalWrite (ucPinCS, 0);
  ucData1 = SPI.transfer (0x10);
  ucData2 = SPI.transfer (0x00);
  digitalWrite (ucPinCS, 1);
  return (((ucData1 & 0x0F) << 8) | ucData2);
}

unsigned int Read_ADC_3 () {
  digitalWrite (ucPinCS, 0);
  ucData1 = SPI.transfer (0x18);
  ucData2 = SPI.transfer (0x00);
  digitalWrite (ucPinCS, 1);
  return (((ucData1 & 0x0F) << 8) | ucData2);
}

// ########## NETZWERK SCAN ##########

void Network_Scan () {
  int iNetwrokNum;
  int iCnt;
  String sSSID;
  int iCH;
  int iRSSI;
  unsigned char ucEncryptionType;
  unsigned char* ucBSSID;
  bool isHidden;

  iNetwrokNum = WiFi.scanNetworks (false, true);
  for (iCnt = 0; iCnt < iNetwrokNum; iCnt++) {
   WiFi.getNetworkInfo (iCnt, sSSID, ucEncryptionType, iRSSI, ucBSSID, iCH, isHidden);
   Serial.printf ("%d|%s|Ch:%d|%d dBm|%s|%s\r\n", iCnt + 1, sSSID.c_str(), iCH, iRSSI, ucEncryptionType == ENC_TYPE_NONE ? "open" : "closed", isHidden ? "hidden" : "visible");
  }
}

// ########## SSID SETZEN ##########

void Network_SSID (char *cSSID) {
  sprintf (sSSID, "%s", cSSID);
}

// ########## PASSWORT SETZEN ##########

void Network_Password (char *cPW) {
  sprintf (sPassword, "%s", cPW);
}

// ########## PIN LESEN ##########

void Read_Network_PIN (char *sPIN, unsigned char ucPrint) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  for (iCnt = 0; iCnt < EE_PIN_LEN -1; iCnt++) {
    ucData = EEPROM.read (EE_PIN_ADR + iCnt);
    sPIN[iCnt] = ucData;
    ucSum += ucData;
  }
  sPIN[iCnt] = 0;
  ucData = EEPROM.read (EE_PIN_ADR + iCnt);
  if (ucSum != ucData) {
    sprintf (sPIN, "*0000#");
  }
  if (ucPrint) {
    Serial.printf ("%s\r\n", sPIN);
  }
}

// ########## PIN SCHREIBEN ##########

void Write_Network_PIN (char *cPIN) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  for (iCnt = 0; iCnt < EE_PIN_LEN -1; iCnt++) {
    ucData = cPIN[iCnt];
    ucSum += ucData;
    EEPROM.write (EE_PIN_ADR + iCnt, ucData);
  }
  EEPROM.write (EE_PIN_ADR + iCnt, ucSum);
  EEPROM.commit ();
}

// ########## NETZWERK VERBINDEN DHCP ##########

void Network_Connect () {
int iCnt;
unsigned char ucData;
unsigned char ucSum;

  // Feste IP ungültig machen
  ipAdr[0] = 0;
  ipAdr[1] = 0;
  ipAdr[2] = 0;
  ipAdr[3] = 0;

  WiFi.disconnect ();
  delay (3000);
  WiFi.begin (sSSID, sPassword);

// SSID und Passwort im EEPROM ablegen
  ucSum = 0xAA;
  for (iCnt = 0; iCnt < EE_SSID_SIZE -1; iCnt++) {
    ucData = sSSID[iCnt];
    ucSum += ucData;
    EEPROM.write (EE_SSID_ADR + iCnt, ucData);
  }
  EEPROM.write (EE_SSID_ADR + iCnt, ucSum);

  ucSum = 0xAA;
  for (iCnt = 0; iCnt < EE_PASSWORD_SIZE -1; iCnt++) {
    ucData = sPassword[iCnt];
    ucSum += ucData;
    EEPROM.write (EE_PASSWORD_ADR + iCnt, ucData);
  }
  EEPROM.write (EE_PASSWORD_ADR + iCnt, ucSum);

  EEPROM.commit ();
}

// ########## NETZWERK AUF FESTE IPS SETZEN ##########

void Fixed_IPs () {
unsigned char ucPos;
unsigned char ucPnt;
unsigned char ucCnt;
int iVal[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
char cVal[USART_BUFFER_SIZE];
int iCnt;
unsigned char ucData;
unsigned char ucSum;

  WiFi.disconnect ();
  delay (3000);

  // Parameter mit den IPs
  ucPos = ucOffsetPointer;
  ucPnt = 0;
  ucCnt = 0;
  while (ucOffsetPointer > 0) {
    if (cBuffer[ucPos] >= '0' && cBuffer[ucPos] <= '9') {
      cVal[ucPnt] = cBuffer[ucPos];
      ucPnt++;
    }
    else {
      cVal[ucPnt] = 0;
      ucPnt = 0;
      iVal[ucCnt] = String_2_Value (cVal);
      ucCnt++;
    }
    if (ucCnt >= 16) {
      ipAdr[0] = iVal[0]; 
      ipAdr[1] = iVal[1]; 
      ipAdr[2] = iVal[2]; 
      ipAdr[3] = iVal[3]; 
      ipGateway[0] = iVal[4];
      ipGateway[1] = iVal[5];
      ipGateway[2] = iVal[6];
      ipGateway[3] = iVal[7];
      ipSubnetMask[0] = iVal[8];
      ipSubnetMask[1] = iVal[9];
      ipSubnetMask[2] = iVal[10];
      ipSubnetMask[3] = iVal[11];
      ipDNS[0] = iVal[12];
      ipDNS[1] = iVal[13];
      ipDNS[2] = iVal[14];
      ipDNS[3] = iVal[15];
      WiFi.config (ipAdr, ipDNS, ipGateway, ipSubnetMask);

      // IP-Adressen im EEPROM ablegen
      ucSum = 0xAA;
      for (iCnt = 0; iCnt < 4; iCnt++) {
        EEPROM.write (EE_FIXED_IPS_ADR + iCnt + 0, ipAdr[iCnt]);
        ucSum += ipAdr[iCnt];
      }
      for (iCnt = 0; iCnt < 4; iCnt++) {
        EEPROM.write (EE_FIXED_IPS_ADR + iCnt + 4, ipGateway[iCnt]);
        ucSum += ipGateway[iCnt];
      }
      for (iCnt = 0; iCnt < 4; iCnt++) {
        EEPROM.write (EE_FIXED_IPS_ADR + iCnt + 8, ipSubnetMask[iCnt]);
        ucSum += ipSubnetMask[iCnt];
      }
      for (iCnt = 0; iCnt < 4; iCnt++) {
        EEPROM.write (EE_FIXED_IPS_ADR + iCnt + 12, ipDNS[iCnt]);
        ucSum += ipDNS[iCnt];
      }
      EEPROM.write (EE_FIXED_IPS_ADR + iCnt + 16, ucSum);
      EEPROM.commit ();

      // Neu verbinden
      ucSum = 0xAA;
      for (iCnt = 0; iCnt < EE_SSID_SIZE -1; iCnt++) {
        ucData = EEPROM.read (EE_SSID_ADR + iCnt);
        sSSID [iCnt] = ucData;
        ucSum += ucData;
      }
      sSSID[iCnt] = 0;
      ucData = EEPROM.read (EE_SSID_ADR + iCnt);
      if (ucSum == ucData) {
        ucSum = 0xAA;
        for (iCnt = 0; iCnt < EE_PASSWORD_SIZE -1; iCnt++) {
          ucData = EEPROM.read (EE_PASSWORD_ADR + iCnt);
          sPassword [iCnt] = ucData;
          ucSum += ucData;
        }
        sPassword[iCnt] = 0;
        ucData = EEPROM.read (EE_PASSWORD_ADR + iCnt);
        if (ucSum == ucData) {
          WiFi.begin (sSSID, sPassword);
        }
      }

      break;
    }
    if (cBuffer[ucPos] == 0) {
      break;
    }
    ucPos++;
  }
}

// ########## ZEIT VOM NTP-SERVER HOLEN ##########

void Get_Time_From_NTP (unsigned char ucPrint) {
char sNTPServerName[EE_NTP_NAME_LEN];
IPAddress ipTimeServer;
WiFiUDP hdUDP;
unsigned char sNTPBuffer[NTP_PACKET_SIZE];
int iCnt, iPackLen, iRetry;
unsigned long ulSecSince2000 = 0;
unsigned long ulSec2000Save = 0;
int iNTPCnt = 0;
unsigned char ucGetTime = 0;
unsigned char ucDSTLoop = 0;
int iDaysInYear, iYear;

  if (WiFi.status () == WL_CONNECTED) {
    do {
      Read_NTP_Name (iNTPCnt, sNTPServerName, 0);
      if (sNTPServerName[0] != 'N' && sNTPServerName[1] != 'O' && sNTPServerName[2] != ' ') {
        if (WiFi.hostByName (sNTPServerName, ipTimeServer)) {
          if (ucPrint) {
            Serial.printf ("Connect to NTP-Server: ");
            Serial.println (ipTimeServer);
          }

          // UDP auf Port 2390 hören
          hdUDP.begin (2390);

          // Buffer auf initialisieren
          for (iCnt = 0; iCnt < NTP_PACKET_SIZE; iCnt++) {
            sNTPBuffer[iCnt] = 0;
          }

          // LI = 11 = Clock is unsynchronized
          // VN = 100 = Version 4
          // Mode = 011 = Client
          sNTPBuffer[0] = 0b11100011;
          // Stratum = Unspecified
          sNTPBuffer[1] = 0;
          // Polling Interval
          sNTPBuffer[2] = 6;
          // Peer Clock Precision
          sNTPBuffer[3] = 0xEC;
          // Platzhalter Root Delay & Root Dispersion
          // Reference Identifier
          sNTPBuffer[12]  = 'E';
          sNTPBuffer[13]  = 'S';
          sNTPBuffer[14]  = 'D';
          sNTPBuffer[15]  = 'G';

          // NTP Antwort auf Port 123
          hdUDP.beginPacket (ipTimeServer, 123);
          hdUDP.write (sNTPBuffer, NTP_PACKET_SIZE);
          hdUDP.endPacket ();

          iRetry = 5;
          while (iRetry > 0) {
            iRetry--;
            delay (250);

            iPackLen = hdUDP.parsePacket ();
            if (iPackLen > 0) {
              iRetry = 0;
              ucGetTime = 1;
              hdUDP.read (sNTPBuffer, NTP_PACKET_SIZE);

              // Sekunden umrechen auf das Jahr 2000
              ulSecSince2000 = ((sNTPBuffer[40] << 24) | (sNTPBuffer[41] << 16) | (sNTPBuffer[42] << 8) | sNTPBuffer[43]) - 3155673600;
              ulSecSince2000 += (Read_Time_Zone(0) * 3600);
               ulSec2000Save = ulSecSince2000;
              ucDSTLoop = 2;
              while (ucDSTLoop) {
                RTC.Second = ulSecSince2000 % 60;
                ulSecSince2000 /= 60;
                RTC.Minute = ulSecSince2000 % 60;
                ulSecSince2000 /= 60;
                RTC.Hour = ulSecSince2000 % 24;
                ulSecSince2000 /= 24;
                // 1. Jan 2000 war ein Samstag (+6)
                RTC.WeekDay = ((ulSecSince2000 + 6) % 7);
                iYear = 0;
                // Schaltjahre all 4 Jahre außer 2100, 2200, 2300 und 2400
                while (1) {
                  iDaysInYear = 366;
                  if ((iYear & 3) || iYear == 100 || iYear == 200 || iYear == 300) { iDaysInYear = 365; }
                  if (ulSecSince2000 < iDaysInYear) { break; }
                  ulSecSince2000 -= iDaysInYear;
                  iYear++;
                }
                RTC.Year = iYear + 2000;
                if (iDaysInYear == 365) { ucDaysInMonth[1] = 28; }
                RTC.Month = 0;
                while (1) {
                  if (ulSecSince2000 < ucDaysInMonth[RTC.Month]) {
                    RTC.Day = ulSecSince2000 + 1;
                    break;
                  }
                  ulSecSince2000 -= ucDaysInMonth[RTC.Month];
                  RTC.Month++;
                }

                ucDSTLoop--;
                if (ucDSTLoop) {
                  RTC.DST = 0;
                  if (Is_Daylight_Saving_Time ()) {
                    RTC.DST = 1;
                  }
                  if (Read_DST (0)) {
                    ulSecSince2000 = ulSec2000Save + (RTC.DST * 3600);
                  }
                  else {
                    break; 
                  }
                }
              }

              if (ucPrint) {
                Serial.printf ("%s %0*d-%0*d-%0*d %0*d:%0*d:%0*d\r\n", sWeekName[RTC.WeekDay], 4, RTC.Year, 2, RTC.Month + 1, 2, RTC.Day, 2, RTC.Hour, 2, RTC.Minute, 2, RTC.Second);
              }
              ucIsTimeOK = 1;
            }
            else {
              if (ucPrint) {
                Serial.println ("No NTP packet received");
              }
              ucIsTimeOK = 0;
            }
          }
        }
        else {
          if (ucPrint) {
            Serial.println ("Couldn't resolve IP address");
          }
          ucIsTimeOK = 0;
        }
      }
      iNTPCnt++;
    } while (ucGetTime == 0 && iNTPCnt < 3);
  }
}

// ########## Sommerzeit bestimmen ##########

int Is_Daylight_Saving_Time () {
  if (RTC.Month < 2 || RTC.Month > 9) {
    return 0;
  }

  if (RTC.Day - RTC.WeekDay >= 25 && (RTC.WeekDay || RTC.Hour >= 2)) {
    if (RTC.Month == 9) {
      return 0;
    }
  }
  else {
    if (RTC.Month == 2) {
      return 0;
    }
  }
  return 1;
}

// ########## NTP-SERVER NAME AUS EEPORM LESEN ##########

void Read_NTP_Name (int iAdrOffset, char *sNTPName, unsigned char ucPrint) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  iAdrOffset *= EE_NTP_NAME_LEN;
  for (iCnt = 0; iCnt < EE_NTP_NAME_LEN -1; iCnt++) {
    ucData = EEPROM.read (EE_NTP_NAME_ADR + iAdrOffset + iCnt);
    sNTPName[iCnt] = ucData;
    ucSum += ucData;
  }
  sNTPName[iCnt] = 0;
  ucData = EEPROM.read (EE_NTP_NAME_ADR + iAdrOffset + iCnt);
  if (ucSum != ucData) {
    switch (iAdrOffset) {
      case (EE_NTP_NAME_LEN * 0): sprintf (sNTPName, "0.de.pool.ntp.org"); break;
      case (EE_NTP_NAME_LEN * 1): sprintf (sNTPName, "0.europe.pool.ntp.org"); break;
      case (EE_NTP_NAME_LEN * 2): sprintf (sNTPName, "ptbtime1.ptb.de"); break;
    }
  }
  if (ucPrint) {
    Serial.printf ("%s\r\n", sNTPName);
  }
}

// ########## NTP-SERVER NAME IN EEPORM SCHREIBEN ##########

void Write_NTP_Name (int iAdrOffset, char *cName) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  iAdrOffset *= EE_NTP_NAME_LEN;
  for (iCnt = 0; iCnt < EE_NTP_NAME_LEN -1; iCnt++) {
    ucData = cName[iCnt];
    ucSum += ucData;
    EEPROM.write (EE_NTP_NAME_ADR + iAdrOffset + iCnt, ucData);
  }
  EEPROM.write (EE_NTP_NAME_ADR + iAdrOffset + iCnt, ucSum);
  EEPROM.commit ();
}

// ########## ZEITZONE AUS EEPORM LESEN ##########

int Read_Time_Zone (unsigned char ucPrint) {
char cSum = 0xAA;
char cData;
int iTimeZone;

  iTimeZone = (int)EEPROM.read (EE_TIME_ZONE_ADR);
  cData = EEPROM.read (EE_TIME_ZONE_ADR + 1);
  cSum += iTimeZone;
  if (cSum == cData) {
    iTimeZone -= 128;
    if (ucPrint) {
      Serial.printf ("%d\r\n", iTimeZone);
    }
    return iTimeZone;
  }
  if (ucPrint) {
    Serial.printf ("1\r\n");
  }
  return 1;
}

// ########## ZEITZONE IN EEPORM SCHREIBEN ##########

void Write_Time_Zone (char *cTime) {
int iTimeZone;
char cSum = 0xAA;

  iTimeZone = String_2_Value (cTime) + 128;
  cSum += iTimeZone;
  EEPROM.write (EE_TIME_ZONE_ADR, (char)iTimeZone);
  EEPROM.write (EE_TIME_ZONE_ADR + 1, cSum);
  EEPROM.commit ();
}

// ########## DST AUS EEPORM LESEN ##########

int Read_DST (unsigned char ucPrint) {
char cSum = 0xAA;
char cData;
char cDST;

  cDST = EEPROM.read (EE_DST_ADR);
  cData = EEPROM.read (EE_DST_ADR + 1);
  cSum += cDST;
  if (cSum == cData) {
    if (ucPrint) {
      Serial.printf ("%d\r\n", cDST);
    }
    return cDST;
  }
  if (ucPrint) {
    Serial.printf ("1\r\n");
  }
  return 1;
}

// ########## DST IN EEPORM SCHREIBEN ##########

void Write_DST (char *cValue) {
char cDST = 0;
char cSum = 0xAA;

  if (cValue[0] == '1') {
    cDST = 1;
  }
  cSum += cDST;
  EEPROM.write (EE_DST_ADR, cDST);
  EEPROM.write (EE_DST_ADR + 1, cSum);
  EEPROM.commit ();
}

// ########## GERÄTENAME AUS EEPORM LESEN ##########

void Read_Device_Name (unsigned char ucPrint) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  for (iCnt = 0; iCnt < EE_DEVICE_NAME_LEN -1; iCnt++) {
    ucData = EEPROM.read (EE_DEVICE_NAME_ADR + iCnt);
    sDeviceName[iCnt] = ucData;
    ucSum += ucData;
  }
  sDeviceName[iCnt] = 0;
  ucData = EEPROM.read (EE_DEVICE_NAME_ADR + iCnt);
  if (ucSum != ucData) {
    sprintf (sDeviceName, "NO NAME");
  }
  if (ucPrint) {
    Serial.printf ("%s\r\n", sDeviceName);
  }
}

// ########## GERÄTENAME IN EEPORM SCHREIBEN ##########

void Write_Device_Name (char *cName) {
int iCnt;
unsigned char ucData;
unsigned char ucSum = 0xAA;

  for (iCnt = 0; iCnt < EE_DEVICE_NAME_LEN -1; iCnt++) {
    ucData = cName[iCnt];
    ucSum += ucData;
    EEPROM.write (EE_DEVICE_NAME_ADR + iCnt, ucData);
  }
  EEPROM.write (EE_DEVICE_NAME_ADR + iCnt, ucSum);
  EEPROM.commit ();
}

// ########## STRING ZU ZAHL WANDELN ##########

int String_2_Value (char *cString) {
int iValue = 0;
int iMulti = 1;
int iCnt;

  for (iCnt = strlen (cString) -1; iCnt >= 0; iCnt--) {
    if (cString[iCnt] >= '0' && cString[iCnt] <= '9') {
      iValue += (cString[iCnt] - 48) * iMulti;
      iMulti *= 10;
    }
    else if (cString[iCnt] == '-') {
      iValue *= -1;
    }
  }
  return iValue;
}

// ########## AUFZEICHNUNG DER EREIGNISSE ##########

void Write_Event_Log () {
int iSum = 0xAAAA;
char cSum = 0xAA;
char cPnt;
long lPnt;

  cPnt = EEPROM.read (EE_EVENT_PNT_ADR);
  cSum += cPnt;
  if (EEPROM.read (EE_EVENT_PNT_ADR + 1) != cSum) {
    cPnt = 0;
  }

  if ((iLEDrdSum > 0 || iLEDblSum > 0 || iLEDyeSum > 0) || (RTC.Minute == 0 && RTC.Year > 2000)) {
    lPnt = (long)cPnt;
    lPnt <<= 4;
    cPnt++;
    if (cPnt == EE_EVENT_LOG_SIZE) {
      cPnt = 0;
    }
    // Pointer im EEPROM ablegen
    EEPROM.write (EE_EVENT_PNT_ADR, cPnt);
    cSum = 0xAA;
    cSum += cPnt;
    EEPROM.write (EE_EVENT_PNT_ADR + 1, cSum);
    // Protokoll im EEPROM ablegen
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  0, RTC.WeekDay);
    iSum += RTC.WeekDay;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  1, (char)(RTC.Year >> 8));
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  2, (char)RTC.Year);
    iSum += RTC.Year;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  3, (RTC.Month + 1));
    iSum += (RTC.Month + 1);
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  4, RTC.Day);
    iSum += RTC.Day;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  5, RTC.Hour);
    iSum += RTC.Hour;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  6, RTC.Minute);
    iSum += RTC.Minute;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  7, (char)iLEDrdSum);
    iSum += iLEDrdSum;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  8, (char)iLEDblSum);
    iSum += iLEDblSum;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt +  9, (char)iLEDyeSum);
    iSum += iLEDyeSum;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 10, (char)(iTM >> 8));
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 11, (char)iTM);
    iSum += iTM;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 12, (char)(iRH >> 8));
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 13, (char)iRH);
    iSum += iRH;
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 14, (char)(iSum >> 8));
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt + 15, (char)iSum);
    EEPROM.commit ();
    // Event Zähler Reset
    iLEDrdSum = 0;
    iLEDblSum = 0;
    iLEDyeSum = 0;
  }
}

// ########## AUFZEICHNUNG LÖSCHEN ##########

void Reset_Event_Log () {
int iCnt;
long lPnt;

  // Pointer auf 0 setzen
  EEPROM.write (EE_EVENT_PNT_ADR, 0);
  EEPROM.write (EE_EVENT_PNT_ADR + 1, 0xAA);
  // Log löschen
  for (iCnt = 0; iCnt < EE_EVENT_LOG_SIZE; iCnt++) {
    EEPROM.write (EE_EVENT_LOG_ADR + lPnt, '#');
    lPnt += 16;
  }
  EEPROM.commit ();
}

// ########## EVENT LOG POINTER AUSLESEN ##########

char Read_Event_Log_Pointer (unsigned char ucPrint) {
char cSum = 0xAA;
char cPnt;

  cPnt = EEPROM.read (EE_EVENT_PNT_ADR);
  cSum += cPnt;
  if (EEPROM.read (EE_EVENT_PNT_ADR + 1) == cSum) {
    if (ucPrint) {
      Serial.printf ("%0*d\r\n", 3, cPnt);
    }
    return cPnt;
  }
  if (ucPrint) {
    Serial.printf ("000\r\n");
  }
  return 0;
}

// ########## EREIGNISS AUSLESEN ##########

void Read_Event_Log (unsigned char ucPrint, long lPnt) {
int iSum = 0xAAAA;
int iChkSum;
unsigned char ucLogWDay, ucLogMonth, ucLogDay, ucLogHour, ucLogMinute, ucLogrd, ucLogbl, ucLogye;
int iLogYear, iLogTM, iLogRH;
char sTM[5];
char sRH[5];

  lPnt <<= 4;
  ucLogWDay =   EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  0);
  iSum += ucLogWDay;
  iLogYear =   (EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  1) << 8) | EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  2);
  iSum += iLogYear;
  ucLogMonth =  EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  3);
  iSum += ucLogMonth;
  ucLogDay =    EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  4);
  iSum += ucLogDay;
  ucLogHour =   EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  5);
  iSum += ucLogHour;
  ucLogMinute = EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  6);
  iSum += ucLogMinute;
  ucLogrd =     EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  7);
  iSum += ucLogrd;
  ucLogbl =     EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  8);
  iSum += ucLogbl;
  ucLogye =     EEPROM.read (EE_EVENT_LOG_ADR + lPnt +  9);
  iSum += ucLogye;
  iLogTM =     (EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 10) << 8) | EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 11);
  iSum += iLogTM;
  iLogRH =     (EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 12) << 8) | EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 13);
  iSum += iLogRH;
  iChkSum =    (EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 14) << 8) | EEPROM.read (EE_EVENT_LOG_ADR + lPnt + 15);

  if (iSum == iChkSum) {
    sprintf (sTM, "%0*d", 3, iLogTM);
    sprintf (sRH, "%0*d", 3, iLogRH);
    sprintf (sLogString, "%s %0*d-%0*d-%0*d %0*d:%0*d|%0*d|%0*d|%0*d|%c%c.%c|%c%c.%c", sWeekName[ucLogWDay], 4, iLogYear, 2, ucLogMonth, 2, ucLogDay, 2, ucLogHour, 2, ucLogMinute, 3, ucLogrd, 3, ucLogbl, 3, ucLogye, sTM[0], sTM[1], sTM[2], sRH[0], sRH[1], sRH[2]);
  }
  else {
    sprintf (sLogString, "## ####-##-## ##:##|###|###|###|##.#|##.#");
  }
  if (ucPrint) {
    Serial.printf ("%s\r\n", sLogString);
  }
}
